import sys
from typing import Iterator

from biofiles.common import Reader
from biofiles.types.sequence import SequenceDescription


class FAIReader(Reader):
    def __iter__(self) -> Iterator[SequenceDescription]:
        for line in self._input:
            sequence_id, length_str, byte_offset_str, _, _ = line.rstrip("\n").split(
                "\t"
            )
            yield SequenceDescription(
                id=sequence_id, length=int(length_str), byte_offset=int(byte_offset_str)
            )


if __name__ == "__main__":
    for path in sys.argv[1:]:
        with FAIReader(path) as reader:
            for seq_desc in reader:
                print(seq_desc)



from dataclasses import dataclass

from biofiles.common import Strand


__all__ = ["Repeat"]


@dataclass(frozen=True)
class Repeat:
    sw_score: int
    divergence_percent: float
    deletion_percent: float
    insertion_percent: float
    sequence_id: str
    sequence_start_original: int
    sequence_end_original: int
    sequence_start_c: int
    sequence_end_c: int
    sequence_left: int
    strand: Strand
    repeat_name: str
    repeat_class: str
    repeat_family: str | None
    repeat_start_original: int
    repeat_end_original: int
    repeat_start_c: int
    repeat_end_c: int
    repeat_left: int
    repeat_id: str | None



from collections import deque, defaultdict
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Iterator, TextIO, Type

from biofiles.common import Strand, Reader
from biofiles.types.feature import (
    Feature,
    FeatureMetaclass,
    Relation,
    get_composite_field,
    Dialect,
)


@dataclass(slots=True)
class FeatureDraft:
    idx: int
    sequence_id: str
    source: str
    type_: str
    start_original: int
    end_original: int
    start_c: int
    end_c: int
    score: float | None
    strand: Strand | None
    phase: int | None
    attributes: dict[str, str | list[str]]

    class_: Type[Feature] | None = None
    id: Any = None
    finalized: Feature | None = None


class FeatureTypes:
    ambiguous_type_mapping: dict[str, list[FeatureMetaclass]]
    unique_type_mapping: dict[str, FeatureMetaclass]

    def __init__(self, feature_types: list[FeatureMetaclass]) -> None:
        self.ambiguous_type_mapping = defaultdict(list)
        self.unique_type_mapping = {}

        for ft in feature_types:
            for type in ft.__filter_type__:
                self.ambiguous_type_mapping[type].append(ft)

        for key, fts in [*self.ambiguous_type_mapping.items()]:
            if len(fts) == 1:
                self.unique_type_mapping[key] = fts[0]
                del self.ambiguous_type_mapping[key]
                continue
            self.ambiguous_type_mapping[key] = _sort_by_filter_specificity(fts)


def _sort_by_filter_specificity(fts: list[FeatureMetaclass]) -> list[FeatureMetaclass]:
    """Sort feature classes by their filter specificity, most specific -> least specific."""
    key = lambda ft: bool(ft.__filter_starts__) + bool(ft.__filter_ends__)
    return sorted(fts, key=key, reverse=True)


@dataclass(slots=True)
class FeatureDrafts:
    feature_types: FeatureTypes
    drafts: list[FeatureDraft] = field(default_factory=deque)
    by_class_and_id: dict[tuple[type, Any], FeatureDraft] = field(default_factory=dict)

    def add(self, draft: FeatureDraft) -> None:
        self.drafts.append(draft)
        if class_ := self.feature_types.unique_type_mapping.get(draft.type_.lower()):
            draft.class_ = class_
            draft.id = get_composite_field(
                draft.attributes, class_.__id_attribute_source__
            )
            self.register(draft)

    def register(self, draft: FeatureDraft) -> None:
        if draft.id is None:
            return
        if (key := (draft.class_, draft.id)) in self.by_class_and_id:
            raise ValueError(
                f"duplicate feature ID {draft.id} for class {draft.class_.__name__}"
            )
        self.by_class_and_id[key] = draft


class RawFeatureReader(Reader):
    def __init__(self, input_: TextIO | Path) -> None:
        super().__init__(input_)

    def __iter__(self) -> Iterator[FeatureDraft]:
        raise NotImplementedError


class FeatureReader(Reader):

    def __init__(self, input_: TextIO | Path | str, dialect: Dialect) -> None:
        super().__init__(input_)
        self._feature_types = FeatureTypes(dialect.feature_types)
        self._raw_reader = self._make_raw_feature_reader()

    def _make_raw_feature_reader(self) -> RawFeatureReader:
        raise NotImplementedError

    def __iter__(self) -> Iterator[Feature]:
        fds = FeatureDrafts(self._feature_types)
        for draft in self._raw_reader:
            fds.add(draft)
        yield from self._finalize_drafts(fds)

    def _finalize_drafts(self, fds: FeatureDrafts) -> Iterator[Feature]:
        self._choose_classes(fds)
        self._instantiate_objects(fds)
        self._fill_relations(fds)
        for fd in fds.drafts:
            yield fd.finalized

    def _choose_classes(self, fds: FeatureDrafts) -> None:
        for fd in fds.drafts:
            if fd.class_:
                continue

            fts = self._feature_types.ambiguous_type_mapping[fd.type_]
            matching_fts = [ft for ft in fts if self._check_filters(fds, fd, ft)]
            if not matching_fts:
                raise ValueError(
                    f"no matching classes (out of {len(fts)}) for "
                    f"feature with type {fd.type_!r}, attributes {fd.attributes!r}"
                )
            if len(matching_fts) > 1:
                raise ValueError(
                    f"too many matching classes ({len(matching_fts)}) for "
                    f"feature with type {fd.type_!r}, attributes {fd.attributes!r}"
                )
            ft = matching_fts[0]
            fd.class_ = ft
            fd.id = get_composite_field(fd.attributes, ft.__id_attribute_source__)
            fds.register(fd)

    def _instantiate_objects(self, fds: FeatureDrafts) -> None:
        for fd in fds.drafts:
            fd.finalized = fd.class_(
                sequence_id=fd.sequence_id,
                source=fd.source,
                type_=fd.type_,
                start_original=fd.start_original,
                end_original=fd.end_original,
                start_c=fd.start_c,
                end_c=fd.end_c,
                score=fd.score,
                strand=fd.strand,
                phase=fd.phase,
                attributes=fd.attributes,
            )

    def _fill_relations(self, fds: FeatureDrafts) -> None:
        for fd in fds.drafts:
            for relation in fd.class_.__relations__:
                related_id = get_composite_field(
                    fd.attributes, relation.id_attribute_source
                )
                related_class = relation.inverse.class_
                try:
                    related_fd = fds.by_class_and_id[related_class, related_id]
                except KeyError as exc:
                    raise ValueError(
                        f"can't find related {related_class.__name__} {related_id} for {fd.finalized}"
                    ) from exc
                setattr(fd.finalized, relation.attribute_name, related_fd.finalized)
                if relation.inverse.attribute_name is None:
                    pass
                elif relation.inverse.one_to_one:
                    if (
                        getattr(
                            related_fd.finalized, relation.inverse.attribute_name, None
                        )
                        is not None
                    ):
                        raise ValueError(
                            f"too many related {relation.class_.__name__}s for {related_fd.finalized} (at least 2), "
                            f"expected one-to-one relation"
                        )
                    setattr(
                        related_fd.finalized,
                        relation.inverse.attribute_name,
                        fd.finalized,
                    )
                else:
                    getattr(
                        related_fd.finalized, relation.inverse.attribute_name
                    ).append(fd.finalized)

    def _check_filters(
        self, fds: FeatureDrafts, fd: FeatureDraft, ft: FeatureMetaclass
    ) -> bool:
        if r := ft.__filter_starts__:
            related_fd = self._get_related_feature_draft(fds, fd, r)
            if fd.strand != related_fd.strand:
                return False
            if fd.strand == "+" and fd.start_original != related_fd.start_original:
                return False
            if fd.strand == "-" and fd.end_original != related_fd.end_original:
                return False
        if r := ft.__filter_ends__:
            related_fd = self._get_related_feature_draft(fds, fd, r)
            if fd.strand != related_fd.strand:
                return False
            if fd.strand == "+" and fd.end_original != related_fd.end_original:
                return False
            if fd.strand == "-" and fd.start_original != related_fd.start_original:
                return False
        return True

    def _get_related_feature_draft(
        self, fds: FeatureDrafts, fd: FeatureDraft, r: Relation
    ) -> FeatureDraft:
        related_class = r.inverse.class_
        related_id = fd.attributes[r.id_attribute_source]
        try:
            return fds.by_class_and_id[related_class, related_id]
        except KeyError as exc:
            raise ValueError(
                f"can't find related {related_class.__name__} for "
                f"{fd.class_.__name__} with attributes {fd.attributes!r}"
            ) from exc



__all__ = ["GTFReader", "GTFWriter"]

import sys
from pathlib import Path
from typing import Iterator

from biofiles.common import Writer
from biofiles.dialects.detector import detect_dialect
from biofiles.dialects.genomic_base import Gene, Exon, Feature, CDS, UTR
from biofiles.gff import RawGFFReader
from biofiles.utility.feature import FeatureReader, RawFeatureReader, FeatureDraft


class RawGTFReader(RawGFFReader):
    def __iter__(self) -> Iterator[FeatureDraft]:
        yield from self._read_gff3()

    def _parse_attributes(
        self, line: str, attributes_str: str
    ) -> dict[str, str | list[str]]:
        try:
            result: dict[str, str | list[str]] = {}
            for part in attributes_str.strip().strip(";").split(";"):
                k, v = part.strip().split(None, 1)
                v = v.removeprefix('"').removesuffix('"').replace(r"\"", '"')
                if k in result:
                    if not isinstance(result[k], list):
                        result[k] = [result[k]]
                    result[k].append(v)
                else:
                    result[k] = v
            return result
        except ValueError as exc:
            raise ValueError(
                f"failed to parse attribute string {attributes_str!r}: {exc}"
            ) from exc


class GTFReader(FeatureReader):

    def _make_raw_feature_reader(self) -> RawFeatureReader:
        return RawGTFReader(self._input)


class GTFWriter(Writer):
    def write(self, feature: Feature) -> None:
        fields = (
            feature.sequence_id,
            feature.source,
            feature.type_,
            str(feature.start_c + 1),
            str(feature.end_c),
            str(feature.score) if feature.score is not None else ".",
            str(feature.strand) if feature.strand is not None else ".",
            str(feature.phase) if feature.phase is not None else ".",
            "; ".join(
                f'{k} "' + v.replace('"', r"\"") + '"'
                for k, v in feature.attributes.items()
            ),
        )
        self._output.write("\t".join(fields))
        self._output.write("\n")


if __name__ == "__main__":
    for path in sys.argv[1:]:
        dialect = detect_dialect(Path(path))
        with GTFReader(path, dialect=dialect) as r:
            total_features = 0
            annotated_genes = 0
            annotated_exons = 0
            annotated_cds = 0
            annotated_utrs = 0
            parsed_genes = 0
            parsed_exons = 0
            parsed_cds = 0
            parsed_utrs = 0
            for feature in r:
                total_features += 1
                annotated_genes += "gene" in feature.type_.lower()
                annotated_exons += feature.type_ == "exon"
                annotated_cds += feature.type_.lower() == "cds"
                annotated_utrs += "utr" in feature.type_.lower()
                parsed_genes += isinstance(feature, Gene)
                parsed_exons += isinstance(feature, Exon)
                parsed_cds += isinstance(feature, CDS)
                parsed_utrs += isinstance(feature, UTR)
        print(
            f"{path}: {total_features} features, "
            f"{parsed_genes} genes parsed out of {annotated_genes}, "
            f"{parsed_exons} exons parsed out of {annotated_exons}, "
            f"{parsed_cds} CDS parsed out of {annotated_cds}, "
            f"{parsed_utrs} UTRs parsed out of {annotated_utrs}",
            file=sys.stderr,
        )



import sys
from itertools import islice
from pathlib import Path

from biofiles.dialects.gencode import GENCODE_DIALECT
from biofiles.dialects.refseq import REFSEQ_DIALECT
from biofiles.dialects.stringtie import STRINGTIE_DIALECT
from biofiles.types.feature import Dialect
from biofiles.utility.feature import RawFeatureReader


class CantDetectDialect(Exception):
    pass


class DialectDetector:
    def __init__(self, raw_reader: RawFeatureReader, num_samples: int = 1000) -> None:
        self._raw_reader = raw_reader
        self._num_samples = num_samples

    def detect(self) -> Dialect:
        gencode_rows = 0
        refseq_rows = 0
        stringtie_rows = 0
        total_rows = 0
        for fd in islice(self._raw_reader, self._num_samples):
            total_rows += 1
            source = fd.source.lower()
            if source in ("havana", "ensembl"):
                gencode_rows += 1
            elif source in ("bestrefseq", "bestrefseq%2cgnomon", "gnomon", "refseq"):
                refseq_rows += 1
            elif source in ("stringtie",):
                stringtie_rows += 1

        if gencode_rows > 0 and gencode_rows >= 0.9 * total_rows:
            return GENCODE_DIALECT
        if refseq_rows > 0 and refseq_rows >= 0.9 * total_rows:
            return REFSEQ_DIALECT
        if stringtie_rows > 0 and stringtie_rows >= 0.9 * total_rows:
            return STRINGTIE_DIALECT

        raise CantDetectDialect(
            f"of {total_rows} read rows {gencode_rows} look like GENCODE, "
            f"{refseq_rows} look like RefSeq, {stringtie_rows} look like StringTie"
        )


def detect_dialect(path: Path) -> Dialect:
    if path.suffix == ".gtf":
        from biofiles.gtf import RawGTFReader

        raw_reader = RawGTFReader(path)
    elif path.suffix in (".gff", ".gff3"):
        from biofiles.gff import RawGFFReader

        raw_reader = RawGFFReader(path)
    else:
        raise CantDetectDialect(f"unknown file extension {path.suffix}")
    detector = DialectDetector(raw_reader=raw_reader)
    return detector.detect()


if __name__ == "__main__":
    exit_code = 0
    for path_str in sys.argv[1:]:
        path = Path(path_str)
        try:
            dialect = detect_dialect(path)
            print(f"{path}\t{dialect.name}")
        except CantDetectDialect as exc:
            print(f"Failed to detect dialect for {path}: {exc}", file=sys.stderr)
            exit_code = 1
    sys.exit(exit_code)



from dataclasses import dataclass, Field, field as dataclass_field
from enum import Enum
from typing import dataclass_transform, Type, Any, TypeAlias
from uuid import uuid4

from biofiles.common import Strand

Source: TypeAlias = str | tuple[str, ...]


@dataclass
class Relation:
    """Equivalent of SQL foreign key — declarative description
    of a relation between two types of features."""

    id_attribute_source: Source
    """ Name of GTF/GFF attribute(s) which contains related feature ID. """

    inverse: "InverseRelation | None" = None

    class_: type | None = None
    """ Python class for the related feature. """
    attribute_name: str | None = None


@dataclass
class InverseRelation:
    inverse: Relation
    one_to_one: bool
    class_: type | None = None
    attribute_name: str | None = None


def get_composite_field(
    attributes: dict[str, str], source: Source
) -> str | tuple[str, ...] | None:
    if source is None:
        return None
    if isinstance(source, str):
        return attributes[source]
    return tuple(attributes[attribute_name] for attribute_name in source)


@dataclass_transform()
class FeatureMetaclass(type):
    __id_attribute_source__: Source | None
    """ Name of GTF/GFF attribute(s) which contains the type-unique ID. """

    __filter_type__: tuple[str, ...]
    """ Filter by feature type ("gene", "transcript", etc.). """

    __filter_starts__: Relation | None
    """ Filter by start position — feature starts at the same point as related feature. """

    __filter_ends__: Relation | None
    """ Filter by end position — feature ends at the same point as related feature. """

    __relations__: list[Relation]
    """ All direct relations for this type, for faster parsing. """

    def __new__(
        cls,
        name,
        bases,
        namespace,
        type: str | tuple[str, ...] | None = None,
        starts: Field | None = None,
        ends: Field | None = None,
    ):
        result = super().__new__(cls, name, bases, namespace)
        result.__id_attribute_source__ = cls._find_id_attribute_source(namespace)
        result._fill_relation_classes(namespace)
        result._fill_filters(type=type, starts=starts, ends=ends)
        result._fill_slots()
        result._fill_init_method(namespace)

        # TODO generate dataclass-like __init__ method,
        #      keep all relations optional

        return result

    @staticmethod
    def _find_id_attribute_source(namespace) -> str:
        result: str | None = None
        for key, value in namespace.items():
            match value:
                case Field(metadata={"id_attribute_name": id_attribute_source}):
                    if result:
                        raise TypeError(
                            f"should specify exactly one id_field() in class {result.__name__}"
                        )
                    result = id_attribute_source
        return result

    def _fill_relation_classes(cls, namespace) -> None:
        cls.__relations__ = []
        for key, value in namespace.items():
            match value:
                case Field(metadata={"relation": Relation() as r}):
                    r.class_ = cls
                    r.attribute_name = key
                    if key in cls.__annotations__:
                        # TODO handle optionality and forward refs
                        r.inverse.class_ = cls.__annotations__[key]
                    cls.__relations__.append(r)
                case Field(metadata={"relation": InverseRelation() as r}):
                    r.class_ = cls
                    r.attribute_name = key
                    # TODO calculating r.inverse.class_ based on type annotation

    def _fill_filters(
        cls,
        *,
        type: str | tuple[str, ...] | None = None,
        starts: Field | None = None,
        ends: Field | None = None,
    ) -> None:
        if type is not None:
            cls.__filter_type__ = (type,) if isinstance(type, str) else type

        cls.__filter_starts__ = None
        if starts is not None:
            cls.__filter_starts__ = starts.metadata["relation"]

        cls.__filter_ends__ = None
        if ends is not None:
            cls.__filter_ends__ = ends.metadata["relation"]

    def _fill_slots(cls) -> None:
        cls.__slots__ = [
            key
            for ancestor in cls.__mro__[::-1][1:]
            for key in ancestor.__annotations__
        ]

    def _fill_init_method(cls, namespace) -> None:
        default_arguments: list[str] = []
        non_default_arguments: list[str] = []
        assignments: list[str] = []
        globals: dict[str, Any] = {}

        key_to_ancestor: dict[str, Type] = {}
        for ancestor in cls.__mro__[:-1]:
            for key, value in ancestor.__annotations__.items():
                key_to_ancestor.setdefault(key, ancestor)

        for ancestor in cls.__mro__[::-1][1:]:
            for key, value in ancestor.__annotations__.items():
                if key_to_ancestor[key] is not ancestor:
                    # Overridden in a descendant class.
                    continue

                field_value = getattr(cls, key, None)
                argument, assignment = cls._compose_field(
                    key, value, field_value, globals
                )

                if argument and argument.endswith(" = None"):
                    default_arguments.append(argument)
                elif argument:
                    non_default_arguments.append(argument)
                assignments.append(assignment)

        body = "\n    ".join(assignments)
        all_arguments = [*non_default_arguments, *default_arguments]
        source_code = f"def __init__(self, {', '.join(all_arguments)}):\n    {body}"
        locals = {}
        exec(source_code, globals, locals)
        cls.__init__ = locals["__init__"]

    def _compose_field(
        cls,
        field_name: str,
        field_annotation: Any,
        field_value: Field | None,
        globals: dict[str, Any],
    ) -> tuple[str | None, str]:
        argument: str | None
        assignment: str
        match field_value:
            case Field(metadata={"relation": r}):
                argument = f"{field_name}: {cls._format_type_arg(field_annotation, optional=True)} = None"
                if isinstance(r, InverseRelation) and not r.one_to_one:
                    assignment = f"self.{field_name} = {field_name} if {field_name} is not None else []"
                else:
                    assignment = f"self.{field_name} = {field_name}"
            case Field(metadata={"id_attribute_name": None}):
                argument = None
                assignment = f"self.{field_name} = None"
            case Field(metadata={"attribute_name": attribute_name}) | Field(
                metadata={"id_attribute_name": attribute_name}
            ):
                default = field_value.metadata.get("attribute_default", _no_default)
                default_factory = field_value.metadata.get(
                    "attribute_default_factory", _no_default
                )
                default_variable_name = f"default_{uuid4().hex}"
                argument = None
                if isinstance(attribute_name, str):
                    if default is not _no_default:
                        globals[default_variable_name] = default
                        getter = f"attributes.get({repr(attribute_name)}, {default_variable_name})"
                    elif default_factory is not _no_default:
                        globals[default_variable_name] = default_factory
                        getter = f"attributes.get({repr(attribute_name)}, {default_variable_name}())"
                    else:
                        getter = f"attributes[{repr(attribute_name)}]"
                else:
                    if default is not _no_default or default_factory is not _no_default:
                        raise NotImplementedError()
                    globals["get_composite_field"] = get_composite_field
                    getter = f"get_composite_field(attributes, {repr(attribute_name)})"
                if isinstance(field_annotation, type) and issubclass(
                    field_annotation, (int, float)
                ):
                    getter = f"{field_annotation.__name__}({getter})"
                elif isinstance(field_annotation, type) and issubclass(
                    field_annotation, Enum
                ):
                    globals[field_annotation.__name__] = field_annotation
                    getter = f"{field_annotation.__name__}({getter})"
                # TODO int | None, list[Enum], etc.
                # TODO ensure it's a list if annotated as list
                assignment = f"self.{field_name} = {getter}"
                # TODO necessary conversions, proper exceptions
            case None:
                argument = f"{field_name}: {cls._format_type_arg(field_annotation, optional=False)}"
                assignment = f"self.{field_name} = {field_name}"
            case property():
                argument = None
                assignment = ""
            case other:
                raise TypeError(f"unsupported field: {field_value}")
        return argument, assignment

    def _format_type_arg(cls, type: str | Type, optional: bool) -> str:
        if isinstance(type, str):
            return f'"{type} | None"' if optional else type
        try:
            if type.__module__ == "builtins":
                return f"{type.__name__} | None" if optional else type.__name__
            return f'"{type.__module__}.{type.__name__}"'
        except AttributeError:
            # TODO Properly support Optional, Union, etc., especially with built-in types
            return f'"{str(type)} | None"' if optional else repr(str(type))


class Feature(metaclass=FeatureMetaclass):
    sequence_id: str
    source: str
    type_: str

    start_original: int
    end_original: int
    # Original values as they were present in the file (1-based inclusive for .gff and .gtf).

    start_c: int
    end_c: int
    # Standardized ("C-style") 0-based values, start inclusive, end exclusive.

    score: float | None
    strand: Strand | None
    phase: int | None
    attributes: dict[str, str]

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.sequence_id}:{self.start_c}-{self.end_c})"


def id_field(source: Source) -> Field:
    return dataclass_field(metadata={"id_attribute_name": source})


def no_id_field() -> Field:
    return dataclass_field(metadata={"id_attribute_name": None})


_no_default = object()


def field(
    source: Source, *, default: Any = _no_default, default_factory: Any = _no_default
) -> Field:
    metadata = {"attribute_name": source}
    if default is not _no_default:
        metadata["attribute_default"] = default
    if default_factory is not _no_default:
        metadata["attribute_default_factory"] = default_factory
    return dataclass_field(metadata=metadata)


def relation(source: Source, *, one_to_one: bool = False) -> tuple[Field, Field]:
    forward_relation = Relation(id_attribute_source=source)
    inverse_relation = InverseRelation(inverse=forward_relation, one_to_one=one_to_one)
    forward_relation.inverse = inverse_relation

    forward_field = dataclass_field(metadata={"relation": forward_relation})
    inverse_field = dataclass_field(metadata={"relation": inverse_relation})
    return forward_field, inverse_field


@dataclass(frozen=True)
class Dialect:
    name: str
    feature_types: list[Type[Feature]]



from biofiles.types.feature import Feature


class Gene(Feature):
    id: str
    biotype: str
    name: str
    transcripts: list["Transcript"]


class Transcript(Feature):
    id: str
    biotype: str
    gene: Gene
    exons: list["Exon"]


class Exon(Feature):
    number: int
    transcript: Transcript
    cdss: list["CDS"]


class CDS(Feature):
    exon: Exon


class UTR(Feature):
    transcript: Transcript


class FivePrimeUTR(UTR):
    pass


class ThreePrimeUTR(UTR):
    pass



"""Feature dialect for StringTie .gtf/.gff output."""

from biofiles.types.feature import Feature, relation, id_field, field, Dialect

exon_transcript, transcript_exons = relation(source="transcript_id")


class Transcript(Feature, type="transcript"):
    id: str = id_field(source="transcript_id")
    gene_id: str = field(source="gene_id")
    exons: list["Exon"] = transcript_exons
    coverage: float = field(source="cov")
    fpkm: float = field(source="FPKM")
    tpm: float = field(source="TPM")


class Exon(Feature, type="exon"):
    id: tuple[str, int] = id_field(source=("transcript_id", "exon_number"))
    transcript: Transcript = exon_transcript
    coverage: float = field(source="cov")


STRINGTIE_FEATURE_TYPES = [Transcript, Exon]
STRINGTIE_DIALECT = Dialect(name="StringTie", feature_types=STRINGTIE_FEATURE_TYPES)



import sys
from collections import Counter
from typing import Iterator, cast, Literal

from biofiles.common import Reader
from biofiles.types.repeat import Repeat


__all__ = ["RepeatMaskerReader"]


class RepeatMaskerReader(Reader):
    def __iter__(self) -> Iterator[Repeat]:
        has_passed_header = False
        for line in self._input:
            parts = line.split()
            if not (14 <= len(parts) <= 15):
                # Probably some metainfo. No way to tell.
                continue
            if not has_passed_header and ("SW" in parts or "score" in parts):
                continue
            has_passed_header = True

            (
                sw_score_str,
                div_str,
                del_str,
                ins_str,
                seq_id,
                seq_start_str,
                seq_end_str,
                seq_left_str,
                strand_str,
                repeat_name,
                repeat_class_family,
                repeat_start_str,
                repeat_end_str,
                repeat_left_str,
                *repeat_id_or_none,
            ) = parts

            sw_score = int(sw_score_str)
            div_percent = float(div_str)
            del_percent = float(del_str)
            ins_percent = float(ins_str)
            seq_start = int(seq_start_str)
            seq_end = int(seq_end_str)
            seq_left = int(seq_left_str[1:-1])
            strand = cast(Literal["+", "-"], {"+": "+", "C": "-"}[strand_str])

            if "/" in repeat_class_family:
                repeat_class, repeat_family = repeat_class_family.split("/", 1)
            else:
                repeat_class, repeat_family = repeat_class_family, None
            if strand_str == "C":
                repeat_start_str, repeat_left_str = (repeat_left_str, repeat_start_str)
            repeat_start = int(repeat_start_str)
            repeat_end = int(repeat_end_str)
            repeat_left = int(repeat_left_str[1:-1])
            repeat_id = repeat_id_or_none[0] if repeat_id_or_none else None
            yield Repeat(
                sw_score=sw_score,
                divergence_percent=div_percent,
                insertion_percent=ins_percent,
                deletion_percent=del_percent,
                sequence_id=seq_id,
                sequence_start_original=seq_start,
                sequence_end_original=seq_end,
                sequence_start_c=seq_start - 1,
                sequence_end_c=seq_end,
                sequence_left=seq_left,
                strand=strand,
                repeat_name=repeat_name,
                repeat_class=repeat_class,
                repeat_family=repeat_family,
                repeat_start_original=repeat_start,
                repeat_end_original=repeat_end,
                repeat_start_c=repeat_start - 1,
                repeat_end_c=repeat_end,
                repeat_left=repeat_left,
                repeat_id=repeat_id,
            )


if __name__ == "__main__":
    for path in sys.argv[1:]:
        with RepeatMaskerReader(path) as r:
            repeats_per_class = Counter(repeat.repeat_class for repeat in r)
        print(f"Repeat classes in {path}:")
        for k, v in repeats_per_class.most_common():
            print(f"    {k}: {v} repeats")



from biofiles.dialects.gencode import GENCODE_DIALECT
from biofiles.gff import GFFReader


def test_profiling() -> None:
    path = "/Users/saluev/bioinf/genomes/hg38/hg38_genomic.gff3"
    reader = GFFReader(path, GENCODE_DIALECT)
    for _ in reader:
        pass



from dataclasses import dataclass


__all__ = ["Sequence"]


@dataclass(frozen=True)
class Sequence:
    id: str
    description: str
    sequence: str


@dataclass(frozen=True)
class SequenceDescription:
    id: str
    length: int
    byte_offset: int



"""Feature dialect for GENCODE .gtf/.gff3 files."""

from enum import StrEnum
from types import NoneType

from biofiles.dialects.genomic_base import (
    Gene as BaseGene,
    Transcript as BaseTranscript,
    Exon as BaseExon,
    CDS as BaseCDS,
    UTR as BaseUTR,
    FivePrimeUTR as BaseFivePrimeUTR,
    ThreePrimeUTR as BaseThreePrimeUTR,
)
from biofiles.types.feature import (
    Feature,
    id_field,
    field,
    relation,
    no_id_field,
    Dialect,
)


class GeneType(StrEnum):
    ARTIFACT = "artifact"
    IG_C_GENE = "IG_C_gene"
    IG_C_PSEUDOGENE = "IG_C_pseudogene"
    IG_D_GENE = "IG_D_gene"
    IG_D_PSEUDOGENE = "IG_D_pseudogene"
    IG_J_GENE = "IG_J_gene"
    IG_J_PSEUDOGENE = "IG_J_pseudogene"
    IG_PSEUDOGENE = "IG_pseudogene"
    IG_V_GENE = "IG_V_gene"
    IG_V_PSEUDOGENE = "IG_V_pseudogene"
    LNCRNA = "lncRNA"
    MIRNA = "miRNA"
    MISC_RNA = "misc_RNA"
    MT_RRNA = "Mt_rRNA"
    MT_TRNA = "Mt_tRNA"
    PROCESSED_PSEUDOGENE = "processed_pseudogene"
    PROTEIN_CODING = "protein_coding"
    RIBOZYME = "ribozyme"
    RRNA = "rRNA"
    RRNA_PSEUDOGENE = "rRNA_pseudogene"
    SRNA = "sRNA"
    SCRNA = "scRNA"
    SCARNA = "scaRNA"
    SNRNA = "snRNA"
    SNORNA = "snoRNA"
    TEC = "TEC"
    TR_C_GENE = "TR_C_gene"
    TR_C_PSEUDOGENE = "TR_C_pseudogene"
    TR_D_GENE = "TR_D_gene"
    TR_D_PSEUDOGENE = "TR_D_pseudogene"
    TR_J_GENE = "TR_J_gene"
    TR_J_PSEUDOGENE = "TR_J_pseudogene"
    TR_V_GENE = "TR_V_gene"
    TR_V_PSEUDOGENE = "TR_V_pseudogene"
    TRANSCRIBED_PROCESSED_PSEUDOGENE = "transcribed_processed_pseudogene"
    TRANSCRIBED_UNITARY_PSEUDOGENE = "transcribed_unitary_pseudogene"
    TRANSCRIBED_UNPROCESSED_PSEUDOGENE = "transcribed_unprocessed_pseudogene"
    TRANSLATED_PROCESSED_PSEUDOGENE = "translated_processed_pseudogene"
    UNITARY_PSEUDOGENE = "unitary_pseudogene"
    UNPROCESSED_PSEUDOGENE = "unprocessed_pseudogene"
    VAULT_RNA = "vault_RNA"


class TranscriptType(StrEnum):
    ARTIFACT = "artifact"
    IG_C_GENE = "IG_C_gene"
    IG_C_PSEUDOGENE = "IG_C_pseudogene"
    IG_D_GENE = "IG_D_gene"
    IG_D_PSEUDOGENE = "IG_D_pseudogene"
    IG_J_GENE = "IG_J_gene"
    IG_J_PSEUDOGENE = "IG_J_pseudogene"
    IG_PSEUDOGENE = "IG_pseudogene"
    IG_V_GENE = "IG_V_gene"
    IG_V_PSEUDOGENE = "IG_V_pseudogene"
    LNCRNA = "lncRNA"
    MIRNA = "miRNA"
    MISC_RNA = "misc_RNA"
    MT_RRNA = "Mt_rRNA"
    MT_TRNA = "Mt_tRNA"
    PROCESSED_PSEUDOGENE = "processed_pseudogene"
    PROTEIN_CODING = "protein_coding"
    RIBOZYME = "ribozyme"
    RRNA = "rRNA"
    RRNA_PSEUDOGENE = "rRNA_pseudogene"
    SRNA = "sRNA"
    SCRNA = "scRNA"
    SCARNA = "scaRNA"
    SNRNA = "snRNA"
    SNORNA = "snoRNA"
    TEC = "TEC"
    TR_C_GENE = "TR_C_gene"
    TR_C_PSEUDOGENE = "TR_C_pseudogene"
    TR_D_GENE = "TR_D_gene"
    TR_D_PSEUDOGENE = "TR_D_pseudogene"
    TR_J_GENE = "TR_J_gene"
    TR_J_PSEUDOGENE = "TR_J_pseudogene"
    TR_V_GENE = "TR_V_gene"
    TR_V_PSEUDOGENE = "TR_V_pseudogene"
    TRANSCRIBED_PROCESSED_PSEUDOGENE = "transcribed_processed_pseudogene"
    TRANSCRIBED_UNITARY_PSEUDOGENE = "transcribed_unitary_pseudogene"
    TRANSCRIBED_UNPROCESSED_PSEUDOGENE = "transcribed_unprocessed_pseudogene"
    TRANSLATED_PROCESSED_PSEUDOGENE = "translated_processed_pseudogene"
    UNITARY_PSEUDOGENE = "unitary_pseudogene"
    UNPROCESSED_PSEUDOGENE = "unprocessed_pseudogene"
    VAULT_RNA = "vault_RNA"

    # Transcript-specific:
    NON_STOP_DECAY = "non_stop_decay"
    NONSENSE_MEDIATED_DECAY = "nonsense_mediated_decay"
    PROCESSED_TRANSCRIPT = "processed_transcript"
    PROTEIN_CODING_CDS_NOT_DEFINED = "protein_coding_CDS_not_defined"
    PROTEIN_CODING_LOF = "protein_coding_LoF"
    RETAINED_INTRON = "retained_intron"


transcript_gene, gene_transcripts = relation(source="gene_id")
selenocysteine_gene, _ = relation(source="gene_id")
selenocysteine_transcript, _ = relation(source="transcript_id")
exon_transcript, transcript_exons = relation(source="transcript_id")
exon_gene, _ = relation(source="gene_id")
cds_exon, exon_cds = relation(source=("transcript_id", "exon_number"), one_to_one=True)
utr_transcript, transcript_utrs = relation(source="transcript_id")
utr_gene, _ = relation(source="gene_id")
five_prime_utr_transcript, transcript_five_prime_utrs = relation(source="transcript_id")
five_prime_utr_gene, _ = relation(source="gene_id")
three_prime_utr_transcript, transcript_three_prime_utrs = relation(
    source="transcript_id"
)
three_prime_utr_gene, _ = relation(source="gene_id")
start_codon_transcript, transcript_start_codons = relation(source="transcript_id")
start_codon_exon, _ = relation(source=("transcript_id", "exon_number"))
stop_codon_transcript, transcript_stop_codons = relation(source="transcript_id")
stop_codon_exon, _ = relation(source=("transcript_id", "exon_number"))


class Gene(BaseGene, type="gene"):
    id: str = id_field(source="gene_id")
    biotype: GeneType = field(source="gene_type")
    name: str = field(source="gene_name")
    transcripts: list["Transcript"] = gene_transcripts
    tags: list[str] = field(source="tag", default_factory=list)


class Transcript(BaseTranscript, type="transcript"):
    id: str = id_field(source="transcript_id")
    biotype: TranscriptType = field(source="transcript_type")
    name: str = field(source="transcript_name")
    gene: Gene = transcript_gene
    exons: list["Exon"] = transcript_exons
    utrs: list["UTR"] = transcript_utrs
    five_prime_utrs: "list[FivePrimeUTR]" = transcript_five_prime_utrs
    three_prime_utrs: "list[ThreePrimeUTR]" = transcript_three_prime_utrs
    start_codon: "list[StartCodon]" = transcript_start_codons
    stop_codons: "list[StopCodon]" = transcript_stop_codons
    tags: list[str] = field(source="tag", default_factory=list)


class Selenocysteine(
    Feature, type=("selenocysteine", "stop_codon_redefined_as_selenocysteine")
):
    id: str = no_id_field()
    gene: Gene = selenocysteine_gene
    transcript: Transcript = selenocysteine_transcript


class Exon(BaseExon, type="exon"):
    id: tuple[str, int] = id_field(source=("transcript_id", "exon_number"))
    number: int = field(source="exon_number")
    transcript: Transcript = exon_transcript
    gene: Gene = exon_gene
    cds: "CDS | None" = exon_cds
    tags: list[str] = field(source="tag", default_factory=list)

    @property
    def cdss(self) -> list["CDS"]:
        # In RefSeq, exon can have multiple CDS.
        # This property is for compatibility with a more general case.
        return [self.cds] if self.cds is not None else []


class CDS(BaseCDS, type="cds"):
    id: tuple[str, int] = id_field(source=("transcript_id", "exon_number"))
    exon: Exon = cds_exon


class UTR(BaseUTR, type="utr"):
    id: NoneType = no_id_field()
    transcript: Transcript = utr_transcript
    gene: Gene = utr_gene


class FivePrimeUTR(BaseFivePrimeUTR, type="five_prime_utr"):
    id: NoneType = no_id_field()
    transcript: Transcript = five_prime_utr_transcript
    gene: Gene = five_prime_utr_gene


class ThreePrimeUTR(BaseThreePrimeUTR, type="three_prime_utr"):
    id: NoneType = no_id_field()
    transcript: Transcript = three_prime_utr_transcript
    gene: Gene = three_prime_utr_gene


class StartCodon(Feature, type="start_codon"):
    id: tuple[str, int] = id_field(source=("transcript_id", "exon_number"))
    transcript: Transcript = start_codon_transcript
    exon: Exon = start_codon_exon


class StopCodon(Feature, type="stop_codon"):
    id: tuple[str, int] = id_field(source=("transcript_id", "exon_number"))
    transcript: Transcript = stop_codon_transcript
    exon: Exon = stop_codon_exon


GENCODE_FEATURE_TYPES = [
    Gene,
    Transcript,
    Selenocysteine,
    Exon,
    CDS,
    UTR,
    FivePrimeUTR,
    ThreePrimeUTR,
    StartCodon,
    StopCodon,
]
GENCODE_DIALECT = Dialect(name="GENCODE", feature_types=GENCODE_FEATURE_TYPES)



"""Feature dialect for RefSeq .gtf/.gff3 files."""

from enum import StrEnum

from biofiles.dialects.genomic_base import (
    Gene as BaseGene,
    Transcript as BaseTranscript,
    Exon as BaseExon,
    CDS as BaseCDS,
)
from biofiles.types.feature import (
    Feature,
    id_field,
    field,
    relation,
    no_id_field,
    Dialect,
)


class GeneType(StrEnum):
    ANTISENSE_RNA = "antisense_RNA"
    C_REGION = "C_region"
    C_REGION_PSEUDOGENE = "C_region_pseudogene"
    D_SEGMENT = "D_segment"
    D_SEGMENT_PSEUDOGENE = "D_segment_pseudogene"
    J_SEGMENT = "J_segment"
    J_SEGMENT_PSEUDOGENE = "J_segment_pseudogene"
    LNCRNA = "lncRNA"
    MIRNA = "miRNA"
    MISC_RNA = "misc_RNA"
    NCRNA = "ncRNA"
    NCRNA_PSEUDOGENE = "ncRNA_pseudogene"
    OTHER = "other"
    PROTEIN_CODING = "protein_coding"
    PSEUDOGENE = "pseudogene"
    RNASE_MRP_RNA = "RNase_MRP_RNA"
    RNASE_P_RNA = "RNase_P_RNA"
    RRNA = "rRNA"
    SCARNA = "scaRNA"
    SCRNA = "scRNA"
    SNORNA = "snoRNA"
    SNRNA = "snRNA"
    TELOMERASE_RNA = "telomerase_RNA"
    TRANSCRIBED_PSEUDOGENE = "transcribed_pseudogene"
    TRNA = "tRNA"
    VAULT_RNA = "vault_RNA"
    V_SEGMENT = "V_segment"
    V_SEGMENT_PSEUDOGENE = "V_segment_pseudogene"
    Y_RNA = "Y_RNA"


class TranscriptType(StrEnum):
    ANTISENSE_RNA = "antisense_RNA"
    C_GENE_SEGMENT = "C_gene_segment"
    D_GENE_SEGMENT = "D_gene_segment"
    J_GENE_SEGMENT = "J_gene_segment"
    LNC_RNA = "lnc_RNA"
    MIRNA = "miRNA"
    MRNA = "mRNA"
    PRIMARY_TRANSCRIPT = "primary_transcript"
    RNASE_MRP_RNA = "RNase_MRP_RNA"
    RNASE_P_RNA = "RNase_P_RNA"
    RRNA = "rRNA"
    SCARNA = "scaRNA"
    SCRNA = "scRNA"
    SNORNA = "snoRNA"
    SNRNA = "snRNA"
    TELOMERASE_RNA = "telomerase_RNA"
    TRANSCRIPT = "transcript"
    TRNA = "tRNA"
    VAULT_RNA = "vault_RNA"
    V_GENE_SEGMENT = "V_gene_segment"
    Y_RNA = "Y_RNA"


transcript_gene, gene_transcripts = relation(source="gene_id")
exon_transcript, transcript_exons = relation(source="transcript_id")
exon_gene, _ = relation(source="gene_id")
cds_exon, exon_cds = relation(source=("transcript_id", "exon_number"))
start_codon_transcript, transcript_start_codon = relation(
    source="transcript_id", one_to_one=True
)
start_codon_exon, _ = relation(source=("transcript_id", "exon_number"), one_to_one=True)
stop_codon_transcript, transcript_stop_codon = relation(
    source="transcript_id", one_to_one=True
)
stop_codon_exon, _ = relation(source=("transcript_id", "exon_number"), one_to_one=True)


class Gene(BaseGene, type="gene"):
    id: str = id_field(source="gene_id")
    biotype: GeneType = field(source="gene_biotype")
    name: str = field(source="gene")
    synonyms: list[str] = field(source="gene_synonym", default_factory=list)
    transcripts: list["Transcript"] = gene_transcripts


class Transcript(BaseTranscript, type="transcript"):
    id: str = id_field(source="transcript_id")
    biotype: TranscriptType = field(source="transcript_biotype")
    product: str | None = field(source="product", default=None)
    gene: Gene = transcript_gene
    exons: list["Exon"] = transcript_exons
    start_codon: "StartCodon | None" = transcript_start_codon
    stop_codon: "StopCodon | None" = transcript_stop_codon


class Exon(BaseExon, type="exon"):
    id: tuple[str, int] = id_field(source=("transcript_id", "exon_number"))
    number: int = field(source="exon_number")
    transcript: Transcript = exon_transcript
    gene: Gene = exon_gene
    cdss: list["CDS"] = exon_cds


class CDS(BaseCDS, type="cds"):
    id: tuple[str, int] = no_id_field()
    exon: Exon = cds_exon


class StartCodon(Feature, type="start_codon"):
    id: tuple[str, int] = id_field(source=("transcript_id", "exon_number"))
    transcript: Transcript = start_codon_transcript
    exon: Exon = start_codon_exon


class StopCodon(Feature, type="stop_codon"):
    id: tuple[str, int] = id_field(source=("transcript_id", "exon_number"))
    transcript: Transcript = stop_codon_transcript
    exon: Exon = stop_codon_exon


REFSEQ_FEATURE_TYPES = [
    Gene,
    Transcript,
    Exon,
    CDS,
    StartCodon,
    StopCodon,
]
REFSEQ_DIALECT = Dialect(name="RefSeq", feature_types=REFSEQ_FEATURE_TYPES)
